function PREC=SCALE(M,PARAM)
% PREC=SCALE(M)
% PREC=SCALE(M,PARAM)
%
% computes an initial diagonal scaling for M and additional column/row 
% permutation. 
% 
% -------------------------------------------------------------------------
%
% Input
% -----
% M                 input matrix
% PARAM             structure of additional parameters
% PARAM.type        type of matrix
%                   0   general (DEFAULT)
%                   1   Hermitian
%                   2   Hermitian, positive definite
% -------------------------------------------------------------------------
%
% Output
% ------
% PREC              structure of preconditioning information
% PREC.pL,PREC.pR   left and right permutation vectors
% PREC.DL,PREC.DR   left and right diagonal matrices such that DL*M(pL,pR)*DR
%                   is well diagonal scaled. This may or may not lead to a 
%                   diagonal dominant matrix. In any case it is assumed that
%                   all columns/rows have a comparable 1-norm.

% known scalings for special classes of matrices
if nargin==2
   if PARAM.type==1
      PREC=SYMSCALE(M,PARAM);
      return
   elseif PARAM.type==2
      PREC=SPDSCALE(M,PARAM);
      return
   end % if-elseif
end % if
 


% ---------------------   now consider the general case   ---------------------
% system size
n=size(M,1);
nodes=1:n;
% reorder M such that sum_i f(|M(i,i)|) is maximal in M(:,PREC.pR), where
% for any x>=0, f is defined by       /    x,   if x<1/(1+alpha)
%                               f(x)=| 
%                                     \  x+1/2, if x>=1/(1+alpha)
%
alpha=0.9;
beta=1/(1+alpha);
% ------------------------   END general settings   --------------------------

% first step: diagonal scaling
% diagonal scaling such that all columns have the same 1-norm
PREC.DL=spdiags(ones(n,1),0,n,n);
PREC.DR=PREC.DR;
% some initial statistics
for j=nodes
    PREC.DR(j,j)=1/sum(abs(M(:,j)));   
end


% second step: column permutation
% initially: natural ordering
PREC.pR=nodes;
PREC.pL=nodes;


% cost functional on the diagonal entries
for i=nodes
    f(i)=abs(M(i,i)*PREC.DR(i,i));
    if f(i)>=beta
       f(i)=f(i)+1/2;
    end % if
end % for


% first step
% Search for any (2x2) block of type  [m_ii m_ij]  the corresponding column
%                                     [m_ji m_jj]   
% permutation that maximizes m_{i,PREC.pR(i)}^2+m_{j,PREC.pR(j)}^2
check=zeros(1,n);
checki=zeros(1,n);
% flag for all rows checked, no improvement possible anymore
allchecked=0;
while ~allchecked 
      allchecked=-1;
      for i=nodes
          if ~check(i)
             idx=find(M(i,PREC.pR));
             % true column index of the diagonal entry
             icp=PREC.pR(i);
             irp=i;
             if M(irp,icp)==0
                idx=[i idx];
                idxi=1;
             else
                idxi=find(idx==i);
             end % if
             checki(idx)=0;
             checki(i)=-1;
             % check for an off-diagonal entry with modulus at least as big as
             % the diagonal entry
             % flag for all entries checked, no improvement possible here
             allcheckedi=0;
             while ~allcheckedi
                   allcheckedi=-1;
                   for j=1:length(idx)
                       jcp=PREC.pR(idx(j));
                       jrp=idx(j);
                       % Is the corresponding other column still available?
                       if ~checki(idx(j))
                          % can we increase a given monotone measure
			  fij=abs(M(irp,jcp)*PREC.DR(jcp,jcp));
                          fji=abs(M(jrp,icp)*PREC.DR(icp,icp);
                          % if abs(M(jrp,icp)*PREC.DR(icp,icp))^2+fij^2>...
			  %    abs(M(irp,icp)*PREC.DR(icp,icp))^2+fji^2
			  if fij>=beta, fij=fij+1/2; end % if
                          if fji>=beta, fji=fji+1/2; end % if
                          if fij+fji>f(i)+f(idx(j))
                             % interchange column i and idx(j)
                             PREC.pR([i idx(j)])=PREC.pR([idx(j) i]);
                             % in any neighbouring row of i,idx(j) now there
                             % might be a possibility for a new improvement,
                             % unckeck those
                             check(find(M(:,icp)))=0;
                             check(find(M(:,jcp)))=0;
                             allchecked=0;
                             % new column i may have new off-diagonal dominant
                             % contributions
                             checki(find(M(:,jcp)))=0;
                             allcheckedi=0;
                             % currently there are two columns not worth 
                             % looking at them, namely i,idx(j). The latter
                             % will be checked anyway
                             checki(i)=-1;
                             icp=jcp;
                             f(i)=fij;
                             f(idx(j))=fji;
                             % new diagonal entry is not suitable for exchange
                          end % if (abs(M(irp,jcp)) ...
                          checki(idx(j))=-1;
                       end % if ~check(idx(j))
                   end % for j=1:length(idx)
             end % while ~allcheckedi
             check(i)=-1;
          end % if ~check(i)
      end % for i
end % while







